"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinarySubtreeDataResolver = void 0;
const base_1 = require("../../base");
const base_2 = require("../../base");
const base_3 = require("../../base");
const ImplicitTilingError_1 = require("./ImplicitTilingError");
/**
 * A class for resolving the binary data that is associated with
 * a subtree.
 *
 * @internal
 */
class BinarySubtreeDataResolver {
    /**
     * Creates a new `BinarySubtreeData` from the given subtree
     * that was parsed from a subtree JSON file.
     *
     * This will resolve all buffer references that are
     * contained in the subtree JSON.
     *
     * @param subtree - The `Subtree`
     * @param resourceResolver - The `ResourceResolver` that
     * will be used to resolve buffer URIs
     * @returns A promise with the `BinarySubtreeData`
     * @throws An ImplicitTilingError when there was a buffer without
     * a URI (which is not valid when no binary buffer was given),
     * or one of the requested buffers could not be resolved.
     */
    static async resolveFromJson(subtree, resourceResolver) {
        const binarySubtreeData = await BinarySubtreeDataResolver.resolveInternal(subtree, undefined, resourceResolver);
        return binarySubtreeData;
    }
    /**
     * Creates a new `BinarySubtreeData` from the given binary subtree
     * data that was directly read from a ".subtree" file.
     *
     * This will extract the JSON- and binary buffer part of the
     * subtree data, and resolve all buffer references that are
     * contained in the subtree JSON.
     *
     * @param input - The whole buffer of a binary subtree file
     * @param resourceResolver - The `ResourceResolver` that
     * will be used to resolve buffer URIs
     * @returns A promise with the `BinarySubtreeData`
     * @throws An ImplicitTilingError when the subtree JSON could
     * not be parsed, or there was a buffer without a URI
     * and no binary buffer was given, or one of the requested
     * buffers could not be resolved.
     */
    static async resolveFromBuffer(input, resourceResolver) {
        const headerByteLength = 24;
        const jsonByteLength = input.readBigUint64LE(8);
        const binaryByteLength = input.readBigUint64LE(16);
        // Extract the JSON data
        const jsonStartByteOffset = headerByteLength;
        const jsonEndByteOffset = jsonStartByteOffset + Number(jsonByteLength);
        const jsonBuffer = input.subarray(jsonStartByteOffset, jsonEndByteOffset);
        let subtreeJson;
        let subtree;
        try {
            subtreeJson = base_1.Buffers.getJson(jsonBuffer);
            subtree = subtreeJson;
        }
        catch (error) {
            throw new ImplicitTilingError_1.ImplicitTilingError("Could not parse subtree JSON data");
        }
        // Extract the binary buffer
        const binaryStartByteOffset = jsonEndByteOffset;
        const binaryEndByteOffset = binaryStartByteOffset + Number(binaryByteLength);
        const binaryBufferSlice = input.subarray(binaryStartByteOffset, binaryEndByteOffset);
        const binaryBuffer = binaryBufferSlice.length > 0 ? binaryBufferSlice : undefined;
        const binarySubtreeData = await BinarySubtreeDataResolver.resolveInternal(subtree, binaryBuffer, resourceResolver);
        return binarySubtreeData;
    }
    /**
     * A thin wrapper around `BinaryBufferDataResolver.resolve`
     * that obtains the binary buffer structure information from
     * the subtree, resolves it, and returns it as part of
     * the `BinarySubtreeData`
     *
     * @param subtree - The `Subtree`
     * @param binaryBuffer - The binary buffer of the subtree
     * @param resourceResolver - The resource resolver
     * @returns A promise to the resolved binary subtree data
     */
    static async resolveInternal(subtree, binaryBuffer, resourceResolver) {
        const binaryBufferStructure = {
            buffers: subtree.buffers ?? [],
            bufferViews: subtree.bufferViews ?? [],
        };
        let binaryBufferData;
        try {
            binaryBufferData = await base_2.BinaryBufferDataResolver.resolve(binaryBufferStructure, binaryBuffer, resourceResolver);
        }
        catch (error) {
            if (error instanceof base_3.BinaryDataError) {
                const message = `Could not read subtree data: ${error.message}`;
                throw new ImplicitTilingError_1.ImplicitTilingError(message);
            }
            throw error;
        }
        return {
            subtree: subtree,
            binaryBufferStructure: binaryBufferStructure,
            binaryBufferData: binaryBufferData,
        };
    }
}
exports.BinarySubtreeDataResolver = BinarySubtreeDataResolver;
//# sourceMappingURL=BinarySubtreeDataResolver.js.map