"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GltfUtilities = void 0;
const gltf_pipeline_1 = __importDefault(require("gltf-pipeline"));
const base_1 = require("../../base");
const tilesets_1 = require("../../tilesets");
const tilesets_2 = require("../../tilesets");
const GltfPipelineLegacy_1 = require("./GltfPipelineLegacy");
const GltfWeb3dQuantizedAttributes_1 = require("./GltfWeb3dQuantizedAttributes");
const base_2 = require("../../base");
const logger = base_2.Loggers.get("contentProcessing");
/**
 * Internal utility methods related to glTF/GLB data.
 *
 * @internal
 */
class GltfUtilities {
    /**
     * Upgrades the binary glTF (GLB) data in the given buffer.
     *
     * The exact pre- and postconditions of this is are intentionally
     * not specified. But it is mainly used for updating a limited
     * subset of glTF 1.0 binary data to glTF 2.0.
     *
     * @param glbBuffer - The buffer containing the GLB
     * @param options - Options for the upgrade that are passed to
     * `gltf-pipeline`
     * @returns A promise that resolves with the upgraded GLB.
     */
    static async upgradeGlb(glbBuffer, options) {
        const result = await gltf_pipeline_1.default.processGlb(glbBuffer, options);
        return result.glb;
    }
    /**
     * Obtains the `version` field value from the header in the given GLB buffer.
     *
     * @param glbBuffer - The GLB buffer
     * @returns The glTF version
     * @throws TileFormatError If the buffer does not contain enough bytes
     * for a glTF header
     */
    static getGltfVersion(glbBuffer) {
        if (glbBuffer.length < 12) {
            throw new tilesets_1.TileFormatError(`Expected at least 12 bytes, but only got ${glbBuffer.length}`);
        }
        const version = glbBuffer.readUInt32LE(4);
        return version;
    }
    /**
     * Extract the JSON- and binary part from the given GLB buffer.
     *
     * The given buffer may contain glTF 2.0 binary data, or glTF 1.0
     * binary data.
     *
     * Note that this does NOT convert the input data. It only extracts
     * the data, as-it-is.
     *
     * @param glbBuffer - The buffer containing the GLB
     * @returns The JSON- and binary data buffers
     * @throws TileFormatError If the input does not contain valid GLB data.
     */
    static extractDataFromGlb(glbBuffer) {
        const magic = base_1.Buffers.getMagicString(glbBuffer);
        if (magic !== "glTF") {
            throw new tilesets_1.TileFormatError(`Expected magic header to be 'glTF', but found ${magic}`);
        }
        if (glbBuffer.length < 12) {
            throw new tilesets_1.TileFormatError(`Expected at least 12 bytes, but only got ${glbBuffer.length}`);
        }
        const length = glbBuffer.readUInt32LE(8);
        if (length > glbBuffer.length) {
            throw new tilesets_1.TileFormatError(`Header indicates ${length} bytes, but input has ${glbBuffer.length} bytes`);
        }
        const version = glbBuffer.readUInt32LE(4);
        if (version === 1) {
            return GltfUtilities.extractDataFromGlb1(glbBuffer);
        }
        if (version === 2) {
            return GltfUtilities.extractDataFromGlb2(glbBuffer);
        }
        throw new tilesets_1.TileFormatError(`Expected version 1 or 2, but got ${version}`);
    }
    /**
     * Internal method for `extractDataFromGlb`, covering glTF 1.0.
     *
     * @param glbBuffer - The buffer containing the GLB
     * @returns The JSON- and binary data buffers
     * @throws TileFormatError If the input does not contain valid GLB data.
     */
    static extractDataFromGlb1(glbBuffer) {
        const headerLength = 20;
        if (glbBuffer.length < headerLength) {
            throw new tilesets_1.TileFormatError(`Expected at least 20 bytes, but only got ${glbBuffer.length}`);
        }
        const length = glbBuffer.readUInt32LE(8);
        const contentLength = glbBuffer.readUint32LE(12);
        const contentFormat = glbBuffer.readUint32LE(16);
        if (contentFormat !== 0) {
            throw new tilesets_1.TileFormatError(`Expected content format to be 0, but found ${contentFormat}`);
        }
        const contentStart = headerLength;
        const contentEnd = contentStart + contentLength;
        if (glbBuffer.length < contentEnd) {
            throw new tilesets_1.TileFormatError(`Expected at least ${contentEnd} bytes, but only got ${glbBuffer.length}`);
        }
        const contentData = glbBuffer.subarray(contentStart, contentEnd);
        const bodyStart = contentEnd;
        const bodyEnd = length;
        const bodyData = glbBuffer.subarray(bodyStart, bodyEnd);
        return {
            jsonData: contentData,
            binData: bodyData,
        };
    }
    /**
     * Internal method for `extractDataFromGlb`, covering glTF 2.0.
     *
     * @param glbBuffer - The buffer containing the GLB
     * @returns The JSON- and binary data buffers
     * @throws TileFormatError If the input does not contain valid GLB data.
     */
    static extractDataFromGlb2(glbBuffer) {
        if (glbBuffer.length < 20) {
            throw new tilesets_1.TileFormatError(`Expected at least 20 bytes, but only got ${glbBuffer.length}`);
        }
        const length = glbBuffer.readUInt32LE(8);
        // Extract the JSON chunk data
        const jsonChunkHeaderStart = 12;
        const jsonChunkLength = glbBuffer.readUint32LE(jsonChunkHeaderStart);
        const jsonChunkType = glbBuffer.readUint32LE(jsonChunkHeaderStart + 4);
        const expectedJsonChunkType = 0x4e4f534a; // ASCII string for "JSON"
        if (jsonChunkType !== expectedJsonChunkType) {
            throw new tilesets_1.TileFormatError(`Expected chunk type to be ${expectedJsonChunkType}, but found ${jsonChunkType}`);
        }
        const jsonChunkStart = 20;
        const jsonChunkEnd = jsonChunkStart + jsonChunkLength;
        if (glbBuffer.length < jsonChunkEnd) {
            throw new tilesets_1.TileFormatError(`Expected at least ${jsonChunkEnd} bytes, but only got ${glbBuffer.length}`);
        }
        const jsonChunkData = glbBuffer.subarray(jsonChunkStart, jsonChunkEnd);
        // Extract the BIN chunk data
        let binChunkData;
        const binChunkHeaderStart = jsonChunkEnd;
        const chunkHeaderLength = 12;
        // If there are not enough bytes for the BIN chunk header
        // after the end of the JSON chunk, then the BIN data
        // will be an empty buffer
        if (binChunkHeaderStart + chunkHeaderLength > length) {
            binChunkData = Buffer.alloc(0);
        }
        else {
            const binChunkLength = glbBuffer.readUint32LE(binChunkHeaderStart);
            const binChunkType = glbBuffer.readUint32LE(binChunkHeaderStart + 4);
            const expectedBinChunkType = 0x004e4942; // ASCII string for "BIN"
            if (binChunkType !== expectedBinChunkType) {
                throw new tilesets_1.TileFormatError(`Expected chunk type to be ${expectedBinChunkType}, but found ${binChunkType}`);
            }
            const binChunkStart = binChunkHeaderStart + 8;
            const binChunkEnd = binChunkStart + binChunkLength;
            if (glbBuffer.length < binChunkEnd) {
                throw new tilesets_1.TileFormatError(`Expected at least ${binChunkEnd} bytes, but only got ${glbBuffer.length}`);
            }
            binChunkData = glbBuffer.subarray(binChunkStart, binChunkEnd);
        }
        return {
            jsonData: jsonChunkData,
            binData: binChunkData,
        };
    }
    /**
     * Extract the JSON part from the given GLB buffer and return it
     * as a buffer.
     *
     * The given buffer may contain glTF 2.0 binary data, or glTF 1.0
     * binary data.
     *
     * Note that this does NOT convert the input data. It only extracts
     * the data, as-it-is.
     *
     * @param glbBuffer - The buffer containing the GLB
     * @returns The JSON buffer
     * @throws TileFormatError If the input does not contain valid GLB data.
     */
    static extractJsonFromGlb(glbBuffer) {
        const data = GltfUtilities.extractDataFromGlb(glbBuffer);
        return data.jsonData;
    }
    /**
     * Creates a glTF 2.0 binary (GLB) buffer from the given JSON and
     * binary chunk data.
     *
     * This is the reverse of `extractDataFromGlb2`.
     *
     * This is a low-level function for turning JSON- and binary chunk
     * data into a GLB buffer. The caller is responsible for making sure
     * that the given JSON matches the given BIN data (for example, that
     * the given JSON does not involve a buffer byte offset that does
     * not fit the given buffer data)
     *
     * @param jsonData - The JSON data
     * @param binData - The binary chunk data
     * @returns The GLB data
     */
    static createGlb2FromData(jsonData, binData) {
        // Ensure 4-byte-alignment for the jsonData
        let jsonDataPadded = jsonData;
        if (jsonData.length % 4 != 0) {
            const paddingLength = 4 - (jsonData.length % 4);
            const padding = Buffer.from(Array(paddingLength).fill(32));
            jsonDataPadded = Buffer.concat([jsonDataPadded, padding]);
        }
        // Ensure 4-byte-alignment for the binData
        let binDataPadded = binData;
        if (binData.length % 4 != 0) {
            const paddingLength = 4 - (binData.length % 4);
            const padding = Buffer.from(Array(paddingLength));
            binDataPadded = Buffer.concat([binDataPadded, padding]);
        }
        // Create the JSON chunk
        const CHUNK_TYPE_JSON = 0x4e4f534a; // ASCII string "JSON"
        const jsonChunkHeader = Buffer.alloc(8);
        jsonChunkHeader.writeInt32LE(jsonDataPadded.length, 0);
        jsonChunkHeader.writeInt32LE(CHUNK_TYPE_JSON, 4);
        const jsonChunkData = Buffer.concat([jsonChunkHeader, jsonDataPadded]);
        // Create the BIN chunk
        const CHUNK_TYPE_BIN = 0x004e4942; // ASCII string "BIN"
        const binChunkHeader = Buffer.alloc(8);
        binChunkHeader.writeInt32LE(binDataPadded.length, 0);
        binChunkHeader.writeInt32LE(CHUNK_TYPE_BIN, 4);
        const binChunkData = Buffer.concat([binChunkHeader, binDataPadded]);
        // Assemble the GLB data
        const MAGIC_BINARY_GLTF_HEADER = 0x46546c67; // ASCII string "glTF"
        const BINARY_GLTF_VERSION = 2;
        // 12 bytes for header
        // length of JSON + 8 bytes for JSON chunk header
        // length of BIN  + 9 bytes for BIN  chunk header
        const length = 12 + jsonDataPadded.length + 8 + binDataPadded.length + 8;
        const header = Buffer.alloc(12);
        header.writeInt32LE(MAGIC_BINARY_GLTF_HEADER, 0);
        header.writeInt32LE(BINARY_GLTF_VERSION, 4);
        header.writeInt32LE(length, 8);
        const glbData = Buffer.concat([header, jsonChunkData, binChunkData]);
        return glbData;
    }
    /**
     * Given an input buffer containing a binary glTF asset, optimize it
     * using gltf-pipeline with the provided options.
     *
     * This method also performs a few updates of certain legacy
     * features that are specific for the GLB data that is part
     * of I3DM and B3DM. Details are not specified here.
     *
     * @param glbBuffer - The buffer containing the binary glTF.
     * @param options - Options specifying custom gltf-pipeline behavior.
     * @returns A promise that resolves to the optimized binary glTF.
     */
    static async optimizeGlb(glbBuffer, options) {
        options = options ?? {};
        if (!options.customStages) {
            options.customStages = [];
        }
        const customStage = async (gltf) => {
            await GltfPipelineLegacy_1.GltfPipelineLegacy.process(gltf);
            return gltf;
        };
        options.customStages.push(customStage);
        const result = await gltf_pipeline_1.default.processGlb(glbBuffer, options);
        return result.glb;
    }
    /**
     * Given an input buffer containing a binary glTF asset, remove
     * its use of the `CESIUM_RTC` extension by inserting new nodes
     * (above the former root nodes) that contain the RTC center as
     * their translation.
     *
     * @param glbBuffer - The buffer containing the binary glTF.
     * @param gltfUpAxis - The glTF up-axis, defaulting to "Y"
     * @returns A promise that resolves to the resulting binary glTF.
     *
     * @deprecated This uses `gltf-pipeline` to replace the CESIUM_RTC
     * extension, is only applicable to GLB data, and may affect the
     * structure of the GLB in a way that is hard to predict. Use
     * the `replaceCesiumRtcExtensionInGltf2Glb` function or the
     * `replaceCesiumRtcExtensionInGltf` to perform this operation
     * only when necessary, and without other side effects
     */
    static async replaceCesiumRtcExtension(glbBuffer, gltfUpAxis) {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        const customStage = (gltf, options) => {
            GltfUtilities.replaceCesiumRtcExtensionInGltf(gltf, gltfUpAxis);
            return gltf;
        };
        const options = {
            customStages: [customStage],
            keepUnusedElements: true,
        };
        const result = await gltf_pipeline_1.default.processGlb(glbBuffer, options);
        return result.glb;
    }
    /**
     * Replaces the `CESIUM_RTC` extension in the given glTF 2.0 GLB data.
     *
     * If the given glTF does NOT use the `CESIUM_RTC` extension, then
     * the given buffer will be returned unmodified.
     *
     * Otherwise, this will insert a new parent node above each root node
     * of a scene. These new parent nodes will have a `translation`
     * that is derived taken from the `CESIUM_RTC` `center`,
     * possibly adjusted to take the up-axis of the glTF into account.
     *
     * The `CESIUM_RTC` extension object and its used/required
     * usage declarations will be removed.
     *
     * @param glb - The GLB data
     * @param gltfUpAxis - The glTF up-axis, defaulting to "Y"
     * @returns The buffer with the updated glTF data, or the
     * given input buffer if the glTF did not use CESIUM_RTC
     */
    static replaceCesiumRtcExtensionInGltf2Glb(glb, gltfUpAxis) {
        // Examine the glTF JSON to see whether it contains the CESIUM_RTC
        // extension.
        const gltfData = GltfUtilities.extractDataFromGlb(glb);
        const gltfJson = JSON.parse(gltfData.jsonData.toString("utf8"));
        const extensionsUsed = gltfJson.extensionsUsed || [];
        if (!extensionsUsed.includes("CESIUM_RTC")) {
            return glb;
        }
        // When te extension is found, it is converted into a root node
        // translation.
        logger.info("Found CESIUM_RTC in GLB - replacing with root node translation");
        GltfUtilities.replaceCesiumRtcExtensionInGltf(gltfJson, gltfUpAxis);
        const gltfJsonBuffer = Buffer.from(JSON.stringify(gltfJson, null, 2));
        return GltfUtilities.createGlb2FromData(gltfJsonBuffer, gltfData.binData);
    }
    /**
     * Replaces the `CESIUM_RTC` extension in the given glTF 2.0 JSON data.
     *
     * If the given glTF does NOT use the `CESIUM_RTC` extension, then
     * the given buffer will be returned unmodified.
     *
     * Otherwise, this will insert a new parent node above each root node
     * of a scene. These new parent nodes will have a `translation`
     * that is derived taken from the `CESIUM_RTC` `center`,
     * possibly adjusted to take the up-axis of the glTF into account.
     *
     * The `CESIUM_RTC` extension object and its used/required
     * usage declarations will be removed.
     *
     * @param gltfJsonBuffer - The glTF JSON buffer
     * @param gltfUpAxis - The glTF up-axis, defaulting to "Y"
     * @returns The buffer with the updated glTF JSON data, or the
     * given input buffer if the glTF did not use CESIUM_RTC
     */
    static replaceCesiumRtcExtensionInGltf2Json(gltfJsonBuffer, gltfUpAxis) {
        // Examine the glTF JSON to see whether it contains the CESIUM_RTC
        // extension.
        const gltfJson = JSON.parse(gltfJsonBuffer.toString("utf8"));
        const extensionsUsed = gltfJson.extensionsUsed || [];
        if (!extensionsUsed.includes("CESIUM_RTC")) {
            return gltfJsonBuffer;
        }
        // When te extension is found, it is converted into a root node
        // translation.
        logger.info("Found CESIUM_RTC in glTF - replacing with root node translation");
        GltfUtilities.replaceCesiumRtcExtensionInGltf(gltfJson, gltfUpAxis);
        return Buffer.from(JSON.stringify(gltfJson, null, 2));
    }
    /**
     * Replaces the `CESIUM_RTC` extension in the given glTF 2.0 JSON object.
     *
     * If the given glTF does NOT use the `CESIUM_RTC` extension, then
     * nothing will be done.
     *
     * Otherwise, this will insert a new parent node above each root node
     * of a scene. These new parent nodes will have a `translation`
     * that is derived taken from the `CESIUM_RTC` `center`,
     * possibly adjusted to take the up-axis of the glTF into account.
     *
     * The `CESIUM_RTC` extension object and its used/required
     * usage declarations will be removed.
     *
     * @param gltf - The glTF 2.0 JSON object
     * @param gltfUpAxis - The glTF up-axis, defaulting to "Y"
     */
    static replaceCesiumRtcExtensionInGltf(gltf, gltfUpAxis) {
        if (!gltf.extensions) {
            return;
        }
        const rtcExtension = gltf.extensions["CESIUM_RTC"];
        if (!rtcExtension) {
            return;
        }
        // Compute the translation, taking the up-axis transform into account
        let rtcTranslation;
        if (gltfUpAxis === "X") {
            rtcTranslation = [
                rtcExtension.center[1],
                -rtcExtension.center[2],
                -rtcExtension.center[0],
            ];
        }
        else if (gltfUpAxis === "Y" || gltfUpAxis === undefined) {
            rtcTranslation = [
                rtcExtension.center[0],
                rtcExtension.center[2],
                -rtcExtension.center[1],
            ];
        }
        else {
            rtcTranslation = [
                rtcExtension.center[0],
                rtcExtension.center[1],
                rtcExtension.center[2],
            ];
        }
        const scenes = gltf.scenes;
        if (!scenes) {
            return;
        }
        for (const scene of scenes) {
            const sceneNodes = scene.nodes;
            if (sceneNodes) {
                for (let i = 0; i < sceneNodes.length; i++) {
                    const nodeIndex = sceneNodes[i];
                    const newParent = {
                        translation: rtcTranslation,
                        children: [nodeIndex],
                    };
                    const newParentIndex = gltf.nodes.length;
                    gltf.nodes.push(newParent);
                    sceneNodes[i] = newParentIndex;
                }
            }
        }
        tilesets_2.Extensions.removeExtensionUsed(gltf, "CESIUM_RTC");
        tilesets_2.Extensions.removeExtension(gltf, "CESIUM_RTC");
    }
    /**
     * Given an input buffer containing a binary glTF 2.0 asset, remove
     * its use of the `WEB3D_quantized_attributes` extension.
     *
     * See `GltfWeb3dQuantizedAttributes` for further notes about
     * the context where this is used.
     *
     * @param glbBuffer - The buffer containing the binary glTF.
     * @returns A promise that resolves to the resulting binary glTF.
     */
    static async replaceWeb3dQuantizedAttributesExtension(glbBuffer) {
        // Process the GLB data with a custom gltf-pipeline stage that removes
        // the WEB3D_quantized_attributes from the extensionsUsed/Required
        // arrays, and collects the 'decodeMatrix' arrays from the extension
        // objects in the accessors.
        const extensionName = "WEB3D_quantized_attributes";
        let usedExtension = false;
        const decodeMatrices = [];
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        const customStage = (gltf, options) => {
            usedExtension = tilesets_2.Extensions.usesExtension(gltf, extensionName);
            if (usedExtension) {
                tilesets_2.Extensions.removeExtensionUsed(gltf, extensionName);
                // Collect the 'decodeMatrix' arrays, one for each accessor
                // (or 'undefined' if the accessor did not use the extension)
                const accessors = gltf.accessors || [];
                for (let i = 0; i < accessors.length; i++) {
                    const accessor = accessors[i];
                    const extensions = accessor.extensions || {};
                    const extension = extensions[extensionName] || {};
                    const decodeMatrix = extension.decodeMatrix;
                    decodeMatrices.push(decodeMatrix);
                }
            }
            return gltf;
        };
        const options = {
            customStages: [customStage],
            keepUnusedElements: true,
        };
        const result = await gltf_pipeline_1.default.processGlb(glbBuffer, options);
        const preprocessedGlb = result.glb;
        // If the glTF did not use the extension, then just return the result
        if (!usedExtension) {
            return preprocessedGlb;
        }
        // Otherwise, post-process the glTF to dequantize the accessors
        // using the decode matrices that have been found for them.
        return GltfWeb3dQuantizedAttributes_1.GltfWeb3dQuantizedAttributes.replaceWeb3dQuantizedAttributesExtension(preprocessedGlb, decodeMatrices);
    }
}
exports.GltfUtilities = GltfUtilities;
//# sourceMappingURL=GltfUtilities.js.map